/***********************************************************************
 * Copyright (c) 2003 University of Kent.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.cs.kent.ac.uk/projects/ocl/CPL.html
 ***********************************************************************/

/**
 *  @author Octavian Patrascoiu
 */

package uk.ac.kent.cs.ocl.plugin.views;

import java.util.Iterator;
import java.util.List;

import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Widget;
import org.eclipse.ui.part.*;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.*;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.swt.SWT;

import uk.ac.kent.cs.kmf.util.ILog;
import uk.ac.kent.cs.kmf.util.OutputStreamLog;
import uk.ac.kent.cs.kmf.util.Pair;
import uk.ac.kent.cs.ocl20.OclProcessor;
import uk.ac.kent.cs.ocl20.bridge4emf.EmfOclProcessorImpl;
import uk.ac.kent.cs.ocl20.semantics.bridge.Environment;
import uk.ac.kent.cs.ocl20.synthesis.RuntimeEnvironment;

/**
 * This class is the OCL view
 */

public class OCLView extends ViewPart  implements ISelectionListener {
	/** GUI members */
	private Composite editor;
	private TabFolder viewer;
	private OclTabSelectionListener list;
	private Text context;
	private Text expression;
	private Text evaluation;
	private Text generation;
	private Text errors;

	/** OCL members */
	private ILog log;
	private OclProcessor processor;
	private Environment env;
	private RuntimeEnvironment renv;

	/**
	 * The constructor.
	 */
	public OCLView() {
		// Inialize the OCL context
		log = new OutputStreamLog(System.out);
		processor = new EmfOclProcessorImpl(log);
		env = processor.environment();
		renv = processor.runtimeEnvironment();
	}

	/**
	 * This is a callback that will allow us
	 * to create the viewer and initialize it.
	 */
	public void createPartControl(Composite parent) {
		// Initialize the Tab
		viewer = new TabFolder(parent, SWT.NONE);
		list = new OclTabSelectionListener();
		viewer.addSelectionListener(list);

		// Initialize the editor
		editor = new Composite(viewer, SWT.NONE);
		TabItem editorTI = new TabItem(viewer, SWT.NONE);
		editorTI.setControl(editor);
		editorTI.setText("Expression");
		GridLayout layout = new GridLayout(1, true);
		editor.setLayout(layout);

		// Initialize the context
		context = new Text(editor, SWT.SINGLE);
		GridData contextData = new GridData(GridData.FILL_HORIZONTAL);
		context.setLayoutData(contextData);
		context.setEditable(false);
		context.setText("context OclVoid inv :");

		// Initialize EXPRESSION tab
		expression = new Text(editor, SWT.MULTI | SWT.V_SCROLL);
		GridData expressionData = new GridData(GridData.FILL_BOTH);
		expression.setLayoutData(expressionData);

		// Initialize EVALUATION tab
		evaluation = new Text(viewer, SWT.MULTI | SWT.V_SCROLL);
		TabItem evaluationTI = new TabItem(viewer, SWT.NONE);
		evaluationTI.setControl( evaluation );
		evaluationTI.setText("Evaluation");

		// Initialize GENERATION tab
		generation = new Text(viewer, SWT.MULTI | SWT.V_SCROLL);
		TabItem generationTI = new TabItem(viewer, SWT.NONE);
		generationTI.setControl(generation);
		generationTI.setText("Generated");

		// Initialize ERRORS tab
		errors = new Text(viewer, SWT.MULTI | SWT.V_SCROLL);
		TabItem logTI = new TabItem(viewer, SWT.NONE);
		log = new EclipseLog(errors);
		logTI.setControl(errors);
		logTI.setText("Errors");

		// add myself as a global selection listener
		getViewSite().getWorkbenchWindow().getSelectionService().addSelectionListener(this);

		// prime the selection
		selectionChanged(null, getSite().getPage().getSelection());		
	}

	public void dispose() {
		getViewSite().getWorkbenchWindow().getSelectionService().removeSelectionListener(this);
		//viewer.removeSelectionListener(list);
		super.dispose();
	}

	private void showMessage(String message) {
		MessageDialog.openInformation(viewer.getShell(), "OCL Evaluator View", message);
	}

	/**
	 * Passing the focus request to the viewer's control.
	 */
	public void setFocus() {
		viewer.setFocus();
	}

	private String getFullName(EObject obj) {
		if (obj == null) return "";
		if (obj instanceof EClass) {
			EClass cls = (EClass)obj;
			return getFullName(cls.getEPackage())+"::"+cls.getName();
		} else {
			EPackage pkg = (EPackage)obj;
			EPackage sPkg = pkg.getESuperPackage(); 
			if (sPkg == null) return pkg.getName();
			else return getFullName(sPkg)+"::"+pkg.getName();			 
		}
	}
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		if (selection != null && selection instanceof IStructuredSelection) {
			IStructuredSelection sel = (IStructuredSelection) selection;
			if (sel.getFirstElement() instanceof EObject) {
				EObject obj = (EObject) sel.getFirstElement();
				EClass cls = obj.eClass();
				String str = "context " + getFullName(cls) + " inv :";
				context.setText(str);
				EPackage pkg = cls.getEPackage();
				while (pkg.getESuperPackage() != null)
					pkg = pkg.getESuperPackage();
				processor.getModels().add(pkg);
				env = processor.environment("self", cls);
				renv = processor.runtimeEnvironment("self", obj);
			} else {
				context.setText("context OclVoid inv :");
				env = processor.environment();
				renv = processor.runtimeEnvironment();
			}
		}
	}

	class OclTabSelectionListener implements SelectionListener {
		Widget currentSel = null;
		public void widgetSelected(SelectionEvent e) {
			TabFolder f = (TabFolder) e.widget;
			TabItem curr = f.getSelection()[0];
			Widget prev = currentSel;
			currentSel = curr.getControl();
			if (currentSel != null) {
				if (!currentSel.equals(prev)) {
					// Evaluate
					if (currentSel.equals(evaluation)) {
						log.reset();
						String expr = context.getText() + " " + expression.getText();
						evaluation.setText("Evaluating '"+expr+"' ...\n");
						Object res = null;
						try {
							res = processor.evaluate(expr, env, renv, log);
						} catch (Exception e1) {							
							res = "[error]";				
						}
						if (res == null) res = "[error]";
						evaluation.append(res.toString());
						log.finalReport();
					// Generate code
					} else if (currentSel.equals(generation)) {
						log.reset();
						String expr = context.getText() + " " + expression.getText();
						generation.setText("Generating code for '"+expr+"' ...\n");
						String res = "";
						try {
							List list = processor.generate(expr, "", env, log);
							Iterator i = list.iterator();
							while (i.hasNext()) {
								Pair p = (Pair)i.next();
								String codeStr = p.getSecond().toString();
								String resultStr = p.getFirst().toString();
								if (codeStr != null && codeStr.length() != 0 && resultStr != null && resultStr.length() != 0)
									res += codeStr+"\n--- Result in '"+resultStr+"'";
								if (i.hasNext()) res +="\n";
							}
						} catch (Exception e1) {
							res = "[error]";				
						}
						if (res.length() == 0) res = "[error]";
						generation.append(res);
						log.finalReport();
					}
				}
			}
		}

		public void widgetDefaultSelected(SelectionEvent e) {
		}
	}

	class EclipseLog implements ILog {
		/** Constructor */
		public EclipseLog(Text text) {
			this.text = text;
		}

		/** Reset the log content */
		public void reset() {
			text.setText("");
			warNo = 0;
			errNo = 0;
		}

		/** Reset the violations counter */
		public void resetViolations() {
			warNo = 0;
			errNo = 0;
		}
		/** Reset the warnings counter */
		public void resetWarnings() {
			warNo = 0;
		}
		/** Reset the errors counter */
		public void resetErrors() {
			errNo = 0;
		}
		/** tooManyViolations */
		public boolean tooManyViolations() {
			return errNo > errUpperBound || warNo > warUpperBound;
		}
		/** Has violations */
		public boolean hasViolations() {
			return errNo != 0 || warNo != 0;
		}
		/** Has errors */
		public boolean hasErrors() {
			return errNo != 0;
		}
		/** Has warnings */
		public boolean hasWarnings() {
			return warNo != 0;
		}
		/** Get warning count */
		public int getWarnings() {
			return warNo;
		}
		/** Get error count */
		public int getErrors() {
			return errNo;
		}

		/** Handling messages and errors */
		public void reportMessage(String message) {
			if (errNo <= errUpperBound && warNo <= warUpperBound) {
				text.append(message+"\n");
			}
		}
		public void printMessage(String message) {
			if (errNo <= errUpperBound && warNo <= warUpperBound) {
				text.append(message);
			}
		}
		public void reportWarning(String message) {
			warNo++;
			if (warNo <= warUpperBound) {
				text.append("Warning: "+message+"\n");
			}
		}
		public void reportWarning(String message, Exception e) {
			warNo++;
			if (warNo <= warUpperBound) {
				text.append("Warning: "+message);
				if (e != null) text.append(e.toString());
				text.append("\n");
			}
		}
		public void reportError(String message) {
			errNo++;
			if (errNo <= errUpperBound) {
				text.append("Error: "+message+"\n");
			}
		}
		public void reportError(String message, Exception e) {
			errNo++;
			if (errNo <= errUpperBound) {
				text.append("Error: "+message);
				if (e != null) text.append(e.toString());
				text.append("\n");
			}
		}

		/** Print final report */
		public void finalReport() {
			if (errNo <= errUpperBound && warNo <= warUpperBound) {
				text.append("Finished - "+errNo+" error(s) "+warNo+" warning(s)"+"\n");
			} else { 
				if (errNo > errUpperBound) {
					text.append("Finished - Too many errors."+"\n");
				} 
				if (warNo > warUpperBound) {
					text.append("Finished - Too many warnings."+"\n");
				}
			}
		}

		/** Set upper bounds */
		public void setWarUpperBound(int bound) {
			warUpperBound = bound;
		}
		public void setErrUpperBound(int bound) {
			errUpperBound = bound;
		}

		/** Close log */
		public void close() {
		}
   
   		/** Text area */
   		private Text text;
		/** warning/error counters */
		protected int warNo = 0;
		protected int errNo = 0;
		/** waning/error bounds */
		protected int warUpperBound = 800;
		protected int errUpperBound = 800;
	}
}
